from __future__ import annotations

import collections.abc
import os
from functools import partial, wraps
from subprocess import CompletedProcess
from types import ModuleType
from typing import Any, Callable, Sequence

from .exceptions import MykeNotFoundError, NoTasksFoundError, TaskAlreadyRegisteredError
from .globals import MYKE_VAR_NAME, ROOT_TASK_KEY, TASKS
from .io.read import read
from .io.write import write
from .run import run, sh
from .utils import _MykeSourceFileLoader, convert_to_command_string, make_executable


def add_tasks(*args: Callable[..., Any], **kwargs: Callable[..., Any]) -> None:
    kwargs.update({x.__name__: x for x in args})

    kwargs = {
        (ROOT_TASK_KEY if k == ROOT_TASK_KEY else convert_to_command_string(k)): v
        for k, v in kwargs.items()
    }

    for k, v in kwargs.items():
        v_existing: Callable[..., Any] | None = TASKS.get(k, None)
        if v_existing:
            raise TaskAlreadyRegisteredError(
                (
                    f"Failed to import module '{v.__module__}': "
                    f"Task '{k}' already defined from module '{v_existing.__module__}'"
                ),
            )
        TASKS[k] = v


def import_module(*mykefiles: str, overwrite: bool | None = None) -> None:
    n_tasks_before: int = len(TASKS)
    for m in mykefiles:
        if m.startswith("https://"):
            m = install_module(m, overwrite=overwrite)

        loader = _MykeSourceFileLoader(os.path.relpath(m), m)
        mod: ModuleType = ModuleType(loader.name)
        loader.exec_module(mod)

        if not hasattr(mod, MYKE_VAR_NAME):
            raise MykeNotFoundError(m)

        n_tasks_after: int = len(TASKS)
        if n_tasks_after <= n_tasks_before:
            raise NoTasksFoundError(m)
        n_tasks_before = n_tasks_after


def install_module(
    url: str,
    path: str | None = None,
    fail_if_exists: bool | None = None,
    overwrite: bool | None = None,
) -> str:
    if overwrite is None:
        overwrite = bool(os.getenv("MYKE_UPDATE_MODULES"))

    if not url.startswith("https://"):
        raise ValueError("Download URLs must start with 'https://'")

    if path is None:
        path = "tasks"

        if not os.path.exists(path):
            os.mkdir(path)

        path = os.path.join(path, os.path.basename(url))

    if not overwrite and not fail_if_exists and os.path.exists(path):
        return path

    resp_text: str = read.url(url)

    import_myke: str = f"import {MYKE_VAR_NAME}"
    if import_myke not in resp_text:
        raise MykeNotFoundError(f'"{import_myke}" not found response text of {url}')

    write(resp_text, path=path, overwrite=overwrite)

    make_executable(path)

    return path


def task(
    func: Callable[..., Any] | None = None,
    *,
    name: str | None = None,
    root: bool | None = False,
) -> Callable[..., Any] | Callable[..., Callable[..., Any]]:
    if not func:
        return partial(task, name=name, root=root)

    if root:
        name = ROOT_TASK_KEY

    if name:
        add_tasks(**{name: func})
    else:
        add_tasks(func)

    return func


def task_run(
    func: Callable[..., str | Sequence[str]] | None = None,
    *,
    name: str | None = None,
    root: bool | None = False,
    capture_output: bool | None = False,
    echo: bool | None = True,
    check: bool | None = True,
    cwd: str | None = None,
    env: dict[str, str] | None = None,
    env_update: dict[str, str | None] | None = None,
    timeout: float | None = None,
    executable: str | None = None,
) -> (
    Callable[..., CompletedProcess[bytes | str]]
    | Callable[..., Callable[..., CompletedProcess[bytes | str]]]
):
    if not func:
        return partial(
            task_sh,
            name=name,
            root=root,
            capture_output=capture_output,
            echo=echo,
            check=check,
            cwd=cwd,
            env=env,
            env_update=env_update,
            timeout=timeout,
            executable=executable,
        )

    @wraps(func)
    def _inner_func(*args: Any, **kwargs: Any) -> tuple[str | None, str | None, int]:
        assert func

        func_script: str | Sequence[str] = func(*args, **kwargs)

        if not isinstance(func_script, str) and not isinstance(
            func_script,
            collections.abc.Sequence,
        ):
            raise TypeError(
                "Expected a string or list of strings to be returned for `sh` function",
            )

        return run(
            func_script,
            capture_output=capture_output,
            echo=echo,
            check=check,
            cwd=cwd,
            env=env,
            env_update=env_update,
            timeout=timeout,
            executable=executable,
        )

    if not name:
        name = func.__name__

    return task(_inner_func, name=name, root=root)


def task_sh(
    func: Callable[..., str | Sequence[str]] | None = None,
    *,
    name: str | None = None,
    root: bool | None = False,
    capture_output: bool | None = False,
    echo: bool | None = True,
    check: bool | None = True,
    cwd: str | None = None,
    env: dict[str, str] | None = None,
    env_update: dict[str, str | None] | None = None,
    timeout: float | None = None,
    executable: str | None = None,
) -> (
    Callable[..., CompletedProcess[bytes | str]]
    | Callable[..., Callable[..., CompletedProcess[bytes | str]]]
):
    if not func:
        return partial(
            task_sh,
            name=name,
            root=root,
            capture_output=capture_output,
            echo=echo,
            check=check,
            cwd=cwd,
            env=env,
            env_update=env_update,
            timeout=timeout,
            executable=executable,
        )

    @wraps(func)
    def _inner_func(*args: Any, **kwargs: Any) -> tuple[str | None, str | None, int]:
        assert func

        func_script: str = func(*args, **kwargs)

        if not isinstance(func_script, str) and not isinstance(
            func_script,
            collections.abc.Sequence,
        ):
            raise TypeError(
                "Expected a string to be returned for `sh` function",
            )

        return sh(
            func_script,
            capture_output=capture_output,
            echo=echo,
            check=check,
            cwd=cwd,
            env=env,
            env_update=env_update,
            timeout=timeout,
            executable=executable,
        )

    if not name:
        name = func.__name__

    return task(_inner_func, name=name, root=root)



from __future__ import annotations


class MykeNotFoundError(Exception):
    ...


class TaskAlreadyRegisteredError(Exception):
    ...


class NoTasksFoundError(Exception):
    ...


#  class CalledProcessError(subprocess.CalledProcessError):
#      def __str__(self) -> str:
#          if self.returncode and self.returncode < 0:
#              return super().__str__()
#          else:
#              msg: str = (
#                  f"Command '{self.cmd}' returned non-zero exit status {self.returncode}:"
#              )
#              for attr in "stdout", "stderr":
#                  attr_value: str = getattr(self, attr)
#                  if attr_value:
#                      attr_value_trimmed: str = attr_value.rstrip("\n")
#                      if attr_value_trimmed:
#                          msg += f"\n{attr}:\n{attr_value_trimmed}"
#              return msg



from pathlib import Path
from typing import List

import myke


def test_split_and_trim_text():
    in_str: str = "  \n\n one \n\n  two  \n three\n\n\n four five"
    out_list: List[str] = ["one", "two", "three", "four five"]
    assert myke.utils.split_and_trim_text(in_str) == out_list


def test_is_version():
    assert myke.utils.is_version("0.0.1a1.dev0")
    assert myke.utils.is_version("v0.0.1a1.dev0")
    assert not myke.utils.is_version("wat0.0.1a1.dev0")


def test_get_repo_root():
    assert isinstance(myke.utils.get_repo_root(), Path)



from __future__ import annotations

from typing import Callable

from yapx.types import Annotated  # pylint: disable=unused-import # noqa: F401
from yapx.types import Literal  # pylint: disable=unused-import # noqa: F401
from yapx.types import Protocol  # pylint: disable=unused-import # noqa: F401

__all__ = ["Annotated", "Literal", "Protocol", "MykeType"]


class MykeType(Protocol):
    TASKS: dict[str, Callable[..., None]]



from __future__ import annotations

import os
import re
import stat
from importlib.machinery import SourceFileLoader
from pathlib import Path
from shutil import which
from subprocess import DEVNULL, CalledProcessError, CompletedProcess, run
from typing import Any

from yapx.utils import convert_to_command_string

__all__ = [
    "convert_to_command_string",
    "make_executable",
    "is_version",
    "get_repo_root",
]


def split_and_trim_text(txt: str | None) -> list[str]:
    if txt is None:
        return []

    txt = txt.strip()

    if not txt:
        return []

    return [x_trim for x in txt.splitlines() for x_trim in [x.strip()] if x_trim]


def make_executable(file: str) -> None:
    st = os.stat(file)
    os.chmod(file, st.st_mode | stat.S_IEXEC)


def is_version(txt: str) -> bool:
    # https://github.com/pypa/packaging/blob/main/packaging/version.py
    VERSION_PATTERN = r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
"""
    return (
        re.compile(f"^{VERSION_PATTERN}$", flags=re.VERBOSE | re.IGNORECASE).search(txt)
        is not None
    )


def get_repo_root(path: str | Path | None = None) -> Path | None:
    if not which("git"):
        return None

    if path is None:
        path = Path.cwd()
    elif not isinstance(path, Path):
        path = Path(path)

    if path is None:
        path = Path.cwd()
    elif not isinstance(path, Path):
        path = Path(path)

    try:
        run(
            ["git", "status", "--porcelain"],
            stdout=DEVNULL,
            stderr=DEVNULL,
            cwd=path,
            check=True,
        )
    except CalledProcessError:
        return None

    p: CompletedProcess = run(
        ["git", "rev-parse", "--show-toplevel"],
        cwd=path,
        check=True,
        capture_output=True,
        text=True,
    )

    return Path(p.stdout.strip())


class _MykeSourceFileLoader(SourceFileLoader):
    """SourceFileLoader that does not output '__pycache__'"""

    def _cache_bytecode(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError()

    def set_data(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError()



import subprocess
from typing import List

import pytest
from _pytest.capture import CaptureFixture, CaptureResult

import myke


def test_run(capfd: CaptureFixture):
    expected: str = "hello world"
    p: subprocess.CompletedProcess = myke.run(["python", "-c", f"print('{expected}')"])

    captured: CaptureResult = capfd.readouterr()
    assert not captured.err
    assert captured.out.rstrip() == expected

    assert p.returncode == 0
    assert not p.stderr
    assert not p.stdout


def test_run_shell(capfd: CaptureFixture):
    expected: str = "hello world"
    p: subprocess.CompletedProcess = myke.run(f'echo "{expected}"')

    captured: CaptureResult = capfd.readouterr()
    assert not captured.err
    assert captured.out.rstrip() == expected

    assert p.returncode == 0
    assert not p.stderr
    assert not p.stdout


def test_run_no_echo(capfd: CaptureFixture):
    expected: str = "hello world"
    p: subprocess.CompletedProcess = myke.run(f'echo "{expected}"', echo=False)

    captured: CaptureResult = capfd.readouterr()
    assert not captured.err
    assert not captured.out

    assert p.returncode == 0
    assert not p.stderr
    assert not p.stdout


def test_run_capture_no_echo(capfd: CaptureFixture):
    expected: str = "hello world"
    p: subprocess.CompletedProcess = myke.run(
        f'echo "{expected}"',
        capture_output=True,
        text=True,
        echo=False,
    )

    captured: CaptureResult = capfd.readouterr()
    assert not captured.err
    assert not captured.out

    assert p.returncode == 0
    assert not p.stderr
    assert p.stdout.rstrip() == expected


def test_run_check():
    with pytest.raises(subprocess.CalledProcessError):
        myke.run("exit 1")


def test_run_no_check():
    expected: int = 1
    p: subprocess.CompletedProcess = myke.run(f"exit {expected}", check=False)
    assert p.returncode == expected


def test_sh():
    expected: str = "hello"
    cmd: str = f"echo -n {expected}"
    assert myke.sh(cmd, capture_output=True).stdout.decode("utf-8") == expected


def test_sh_stdout_lines():
    expected: List[str] = ["hello", "world"]
    cmd: str = "echo '" + "\\n".join(expected) + "'"
    stdout: List[str] = myke.sh_stdout_lines(cmd)

    assert stdout == expected


def test_sh_stdout():
    expected: List[str] = ["hello", "world"]
    cmd: str = "echo '" + "\\n".join(expected) + "'"
    stdout: str = myke.sh_stdout(cmd)

    assert stdout == "\n".join(expected)


def test_require(capfd: CaptureFixture):
    p: subprocess.CompletedProcess = myke.require(
        pip_args=[
            "--extra-index-url",
            "https://codeberg.org/api/packages/Fresh2dev/pypi/simple",
            "--quiet",
            "--dry-run",
        ],
        mykefiles="0.0.1a3.dev44",
    )

    captured: CaptureResult = capfd.readouterr()
    assert not captured.err
    assert not captured.out

    assert p.returncode == 0
    assert not p.stderr
